package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.generichash.GenericHash
import com.ionspin.kotlin.crypto.util.encodeToUByteArray
import kotlinx.coroutines.channels.ReceiveChannel
import kotlinx.coroutines.flow.Flow
import org.kotlincrypto.hash.sha3.SHA3_256
import org.kotlincrypto.hash.sha3.SHA3_384

private interface StreamProcessor {
    fun update(data: UByteArray)
    fun final(): UByteArray
}

/**
 * Hash support for crypto2. We implement only secure as for the crypto2 publication time functions,
 * and do not include broken algorithms like SHA1 and SHA2, Md5, etc. To calculate hashes use:
 *
 * - [digest] to calculate a hash of the block in memory
 * - [ofChannel] to calculate hash of long or slow data in a channel
 * - [ofFlow] to calculate hash of the data in the [Flow<UByteArray>]
 */
@Suppress("unused")
enum class Hash(
    private val direct: ((UByteArray) -> UByteArray)? = null,
    private val streamProcessor: () -> StreamProcessor,
) {

    Blake2b(
        // direct blacke2 is faster than stream:
        { GenericHash.genericHash(it) }, {
            object : StreamProcessor {
                val state = GenericHash.genericHashInit()
                override fun update(data: UByteArray) {
                    GenericHash.genericHashUpdate(state, data)
                }

                override fun final(): UByteArray =
                    GenericHash.genericHashFinal(state)
            }
        }),
    Sha3_384(
        // direct Keccaak currently is slower
        null,
        {
            object : StreamProcessor {
                val state = SHA3_384()
                override fun update(data: UByteArray) {
                    state.update(data.asByteArray())
                }

                override fun final(): UByteArray = state.digest().asUByteArray()
            }
        }),
    Sha3_256(null,
        {
            object : StreamProcessor {
                val state = SHA3_256()
                override fun update(data: UByteArray) {
                    state.update(data.asByteArray())
                }

                override fun final(): UByteArray = state.digest().asUByteArray()
            }
        }),

    /**
     * Synthetic hash contains of [Blake2b] and [Sha3_384] hashes concatenated in this order. Let calculate
     * compound hash using [ofFlow] or [ofChannel] effectively.
     */
    Sha3AndBlake(null,
        {
            object : StreamProcessor {
                val blake = Blake2b.streamProcessor()
                val sha384 = Sha3_384.streamProcessor()
                override fun update(data: kotlin.UByteArray) {
                    blake.update(data)
                    sha384.update(data)
                }

                override fun final(): UByteArray = blake.final() + sha384.final()
            }
        });

    @Deprecated("will be removed in favor of digest()", ReplaceWith("digest()"))
    fun perform(src: UByteArray): UByteArray = digest(src)

    /**
     * Calculate digest for the in-memory data
     * @param src data to calculate hash digest over
     * @return calculated hash value
     */
    fun digest(src: UByteArray): UByteArray = direct?.invoke(src) ?: streamProcessor().also { it.update(src) }.final()

    /**
     * Collect the flow and return the hash digest of all the data. Let calculate hashes on data
     * that are too big to fit in memory
     */
    suspend fun ofFlow(source: Flow<UByteArray>): UByteArray {
        val sp = streamProcessor()
        source.collect { sp.update(it) }
        return sp.final()
    }

    /**
     * Read all data from the channel and return the hash digest of all the data. Let calculate hashes on data
     * that are too big to fit in memory
     */
    suspend fun ofChannel(source: ReceiveChannel<UByteArray>): UByteArray {
        val sp = streamProcessor()
        for (block in source) sp.update(block)
        return sp.final()
    }

}

private val defaultSuffix1 = "All lay loads on a willing horse".encodeToUByteArray()
private val defaultSuffix2 = "A stitch in time saves nine".encodeToUByteArray()

/**
 * Caclulate [Hash.Blake2b] hash for [src], shortcut for [Hash.Blake2b.digest]
 */
fun blake2b(src: UByteArray): UByteArray = Hash.Blake2b.digest(src)

/**
 * Double linked Blake2b using the default or specified suffix. This should be more hard to
 * brute force.collision attack than just [blake2b]. Note that different suffixes provide different
 * results.
 */
fun blake2b2l(src: UByteArray, suffix: UByteArray = defaultSuffix1): UByteArray =
    blake2b(blake2b(src) + suffix + src)

/**
 * Triple linked [blake2b], even more prone to collision attacks than [blake2b2l].
 */
fun blake2b3l(src: UByteArray): UByteArray = blake2b(blake2b2l(src) + defaultSuffix2 + src)
