package net.sergeych.crypto2

import kotlinx.datetime.Instant
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bipack.BipackDecoder
import net.sergeych.bipack.BipackEncoder
import net.sergeych.bipack.decodeFromBipack

/**
 * Multi-signed data box. Do not use the constructori directly, use [SealedBox.create]
 * instead to create the box and [SealedBox.plus] or [addSeal] to add more signatures
 * with different signing keys.
 * [SealedBox.contains] checks for a specific key signature presence.
 *
 * Signatures, [Seal], incorporate creation time and optional expiration which are
 * also signed and checked upon deserialization.
 *
 * It is serializable and checks integrity __on deserialization__. If any of seals does not
 * match the signed [message], it throws [IllegalSignatureException] _on deserialization_.
 * E.g., if you have it deserialized, it is ok, check it contains all needed keys among
 * signers.
 *
 * __The main constructor is used for deserializing only__. Don't use it directly unless you
 * know what you are doing as it may be dangerous.Use one of the above to create or change it.
 */
@Serializable
class SealedBox(
    val message: UByteArray,
    private val seals: List<Seal>,
    @Transient
    private val checkOnInit: Boolean = true
) {

    @Suppress("unused")
    constructor(message: UByteArray, vararg keys: SigningKey) :
            this(message, keys.map { it.seal(message) } )

    /**
     * If this instance is not signed by a given key, return new instance signed also by this
     * key, or return unchanged (same) object if it is already signed by this key; you
     * _can't assume it always returns a copied object!_
     */
    operator fun plus(key: SigningKey): SealedBox =
        if (key.verifyingKey in this) this
        else SealedBox(message, seals + key.seal(message),false)

    /**
     * Add expiring seal, otherwise use [plus]. Overrides exising seal for [key]
     * if present:
     */
    fun addSeal(key: SigningKey, expiresAt: Instant): SealedBox {
        val filtered = seals.filter { it.publicKey != key.verifyingKey }
        return SealedBox(message, filtered + key.seal(message, expiresAt), false)
    }

    /**
     * Check that it is signed with a specified key.
     */
    operator fun contains(publicKey: VerifyingKey): Boolean {
        return seals.any { it.publicKey == publicKey }
    }

    init {
        if (seals.isEmpty()) throw IllegalArgumentException("there should be at least one seal")
        if (checkOnInit) {
            for( s in seals ) s.verify(message)
        }
    }

    val encoded: UByteArray by lazy { BipackEncoder.encode(this).toUByteArray() }

    companion object {
        /**
         * Create a new instance with a specific data sealed by one or more
         * keys. At least one key is required to disallow providing not-signed
         * instances, e.g. [SealedBox] is guaranteed to be properly sealed when
         * successfully instantiated.
         *
         * @param data a message to sign
         * @param keys a list of keys to sign with, should be at least one key.
         * @throws IllegalArgumentException if keys are not specified.
         */
        fun create(data: UByteArray, vararg keys: SigningKey): SealedBox {
            return SealedBox(data, keys.map { it.seal(data) }, false)
        }

        inline fun <reified T>encode(value: T, vararg keys: SigningKey): UByteArray =
            create(BipackEncoder.encode(value).toUByteArray(), *keys).encoded

        /**
         * Unpack and check sealed box, returning deserialized contained value
         * @throws IllegalSignatureException
         */
        inline fun <reified T>decode(packedSealedBox: UByteArray): T =
            BipackDecoder.decode<SealedBox>(packedSealedBox.asByteArray()).message.decodeFromBipack()

    }
}