package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.box.Box
import com.ionspin.kotlin.crypto.scalarmult.ScalarMultiplication
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bipack.BipackDecoder

/**
 * The secret key used in public-key encryption; it is used to _decrypt_ data encrypted with its
 * public counterpart, see [publicKey].
 */
@Serializable
@SerialName("encs")
class SecretKey(
    override val keyBytes: UByteArray,
    @Transient
    val _cachedPublicKey: PublicKey? = null,
) : DecryptingKey, UniversalKey() {

    @Transient
    override val label: String = "sec"
    /**
     * Decrypt with authentication checks the message which must have [Asymmetric.Message.senderPublicKey] set.
     * Use [decryptWithSenderKey] otherwise. Note that the authenticated encryption is always use, even if
     * the [PublicKey.encryptAnonymousMessage] was used to create a message, if it is successfully decrypted,
     * it is guaranteed that the message was not altered after creation.
     *
     * @throws DecryptionFailedException If the message is tampered (changed after creation) or was not intended for us,
     */
    fun decrypt(message: Asymmetric.Message): UByteArray = message.decrypt(this)


    /**
     * Decrypt using [senderPublicKey] as a sender key (overriding the [Asymmetric.Message.senderPublicKey] if set).
     * See [decrypt] for more.
     */
    fun decryptWithSenderKey(message: Asymmetric.Message, senderPublicKey: PublicKey): UByteArray =
        message.decryptWithSenderKey(senderPublicKey, this)

    @Transient
    private var cachedPublicKey: PublicKey? = _cachedPublicKey

    /**
     * The corresponding public key
     */
    val publicKey: PublicKey by lazy {
        if (cachedPublicKey != null)
            cachedPublicKey!!
        else
            PublicKey(ScalarMultiplication.scalarMultiplicationBase(keyBytes))
                .also { cachedPublicKey = it }
    }

    /**
     * Nonce-based decryption is impossible, it is already included in message
     */
    override fun decryptWithNonce(cipherData: UByteArray, nonce: UByteArray): UByteArray = decrypt(cipherData)

    /**
     * Decrypt without a nonce as edwards curve decryption does not need it
     */
    override fun decrypt(cipherData: UByteArray): UByteArray {
        val message: Asymmetric.Message = BipackDecoder.decode(cipherData.asByteArray())
        return message.decrypt(this)
    }

    override val magic: KeysmagicNumber = KeysmagicNumber.defaultAssymmetric
    override val id: KeyId by lazy { publicKey.id }

    override val nonceBytesLength: Int
        get() = 0

    companion object {
        data class KeyPair(val secretKey: SecretKey, val publicKey: PublicKey)

        /**
         * Generate a new random pair of public and secret keys.
         */
        fun generateKeys(): KeyPair {
            val p = Box.keypair()
            val pk = PublicKey(p.publicKey)
            return KeyPair(SecretKey(p.secretKey, pk), pk)
        }

        fun new(): SecretKey = generateKeys().secretKey

    }
}