package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.signature.Signature
import kotlinx.datetime.Instant
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.utools.now

/**
 * Secret key to sign only
 */
@Serializable
@SerialName("sigs")
class SigningSecretKey(
        override val keyBytes: UByteArray,
        @Transient
        private var cachedPublicKey: VerifyingPublicKey?=null
    ) : UniversalKey(), SigningKey {

    override val verifyingKey: VerifyingPublicKey by lazy {
        cachedPublicKey ?:
            VerifyingPublicKey(Signature.ed25519SkToPk(keyBytes)).also { cachedPublicKey = it }
    }

    override fun sign(message: UByteArray): UByteArray = Signature.detached(message, keyBytes)

    override fun seal(message: UByteArray, expiresAt: Instant?): Seal =
        Seal.create(this, message, now(), expiresAt)

    override val id: KeyId = verifyingKey.id

    override val label: String
        get() = "sig"

    companion object {

        data class SigningKeyPair(val secretKey: SigningSecretKey, val publicKey: VerifyingPublicKey)

        fun generatePair(): SigningKeyPair {
            val p = Signature.keypair()
            val publicKey = VerifyingPublicKey(p.publicKey)
            return SigningKeyPair(SigningSecretKey(p.secretKey, publicKey), publicKey)
        }

        fun new(): SigningSecretKey = generatePair().secretKey

    }
}